from __future__ import annotations

from collections.abc import Iterable, Sequence
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import reduce
from math import ceil
from queue import SimpleQueue
from typing import TYPE_CHECKING, Any, overload

import numpy as np
import numpy.typing as npt
import polars as pl
import tqdm
from scipy.linalg import norm
from scipy.ndimage import gaussian_filter
from sklearn.decomposition import PCA
from umap import UMAP

from ._kde import _TRUNCATE, KDE_2D, _sample_expression, kde_2d_discrete
from ._patching import _patches, n_patches
from ._subslicing import process_coordinates
from ._utils import (
    UMAP_2D_PARAMS,
    UMAP_RGB_PARAMS,
    _calculate_embedding,
    _cosine_similarity,
    _determine_localmax_and_sample,
    _fill_color_axes,
    _gaussian_weighted_neighbors,
    _minmax_scaling,
    _transform_embeddings,
    _weighted_average_expression,
)

if TYPE_CHECKING:
    import pandas as pd
    from anndata import AnnData


class Ovrlp:
    """
    Main analysis class for spatial overlap analysis.

    Parameters
    ----------
    transcripts : polars.DataFrame | pandas.DataFrame
        Transcript information containing coordinates and gene name/id.
    KDE_bandwidth : float, optional
        The bandwidth of the KDE.
    min_distance : float, optional
        Minimum distance for cell typing.
    n_components : int, optional
        Number of components for PCA.
    gene_key : str
        Name of the gene column
    coordinate_keys : collections.abc.Iterable[str]
        Names of the coordinate columns.
    n_workers : int
        Number of threads used in parallel processing.
    dtype : numpy.typing.DTypeLike
        Datatype used for KDE calculations.
    patch_length : int
        Upper bound for size of each patch. (Only relevant for processing)
    umap_kwargs : dict, optional
        Keyword arguments for 2D UMAP embedding.
    cumap_kwargs : dict, optional
        Keyword arguments for 3D UMAP embedding.
    random_state : int | numpy.random.RandomState | None
        Random state used to seed UMAP and PCA.

    Attributes
    ----------
    transcripts : polars.DataFrame
        Transcript information containing coordinates and gene name/id.
    KDE_bandwidth : float
        The bandwidth of the KDE.
    min_distance : int
        Minimum distance between pseudocells (local maxima).
    pseudocells : anndata.AnnData
        Gene expression matrix of the pseudcells.
    signatures : polars.DataFrame
        A dataframe of celltypes x gene signatures used to annotate the UMAP.
    celltype_centers : numpy.ndarray
        The center of gravity of each celltype in the 2D embedding, used for UMAP annotation.
    celltype_assignments : numpy.ndarray
        The assignments of the cell types.
    pca : sklearn.decomposition.PCA
        The PCA object used for the 2D embedding and calculating the VSI score.
    umap_2d : umap.UMAP
        The UMAP object used for the 2D embedding.
    pca_rgb : sklearn.decomposition.PCA
        The PCA object used for the 3D RGB embedding.
    umap_rgb : umap.UMAP
        The UMAP object used for the 3D RGB embedding.
    genes : list
        A list of genes to utilize in the model.
    integrity_map : numpy.ndarray
        The integrity map of the tissue.
    signal_map : numpy.ndarray
        A pixel map of overall signal strength in the tissue, used to mask out low-signal regions.
    dtype : numpy.typing.DTypeLike
        Datatype used for KDE calculations.
    patch_length : int
        Upper bound for size of each patch. (Only relevant for processing)
    n_workers : int
        Number of threads used in parallel processing.
    """

    def __init__(
        self,
        transcripts: pl.DataFrame | pd.DataFrame,
        /,
        KDE_bandwidth: float = 2.5,
        min_distance: float = 8,
        n_components: int = 30,
        *,
        gene_key: str = "gene",
        coordinate_keys: Iterable[str] = ("x", "y", "z"),
        n_workers: int = 1,
        dtype: npt.DTypeLike = np.float32,
        patch_length: int = 500,
        umap_kwargs: dict[str, Any] = UMAP_2D_PARAMS,
        cumap_kwargs: dict[str, Any] = UMAP_RGB_PARAMS,
        random_state: int | np.random.RandomState | None = None,
    ) -> None:
        columns = {gene_key: "gene"} | dict(zip(coordinate_keys, ["x", "y", "z"]))
        if not isinstance(transcripts, pl.DataFrame):
            transcripts = pl.from_pandas(transcripts)
        self.transcripts = transcripts.rename(columns)

        self.KDE_bandwidth = KDE_bandwidth
        self.dtype = dtype

        self.n_workers = n_workers
        self.patch_length = patch_length

        self.min_distance = min_distance

        umap_kwargs = {"random_state": random_state} | umap_kwargs
        cumap_kwargs = {"random_state": random_state} | cumap_kwargs

        if "n_jobs" not in umap_kwargs:
            # umap_kwargs["random_state"] is not necessarily == random_state
            n_jobs = n_workers if umap_kwargs.get("random_state") is None else 1
            umap_kwargs["n_jobs"] = n_jobs
        if "n_jobs" not in cumap_kwargs:
            n_jobs = n_workers if cumap_kwargs.get("random_state") is None else 1
            cumap_kwargs["n_jobs"] = n_jobs

        self.pca = PCA(n_components=n_components, random_state=random_state)
        self.umap_2d = UMAP(**(umap_kwargs | {"n_components": 2}))
        self.pca_rgb = PCA(n_components=3, random_state=random_state)
        self.umap_rgb = UMAP(**(cumap_kwargs | {"n_components": 3}))

    def process_coordinates(self, gridsize: float = 1, **kwargs):
        """
        Process the coordinates of the transcripts dataframe.

        Parameters
        ----------
        gridsize : float, optional
            The size of the pixel grid.
        kwargs
            Other keyword arguments are passed to :py:func:`ovrlpy.process_coordinates`
        """
        self.transcripts = process_coordinates(
            self.transcripts, gridsize=gridsize, **kwargs
        )

    def _expression_threshold(self, n: float = 10, scale: float = 1.1) -> float:
        return n * scale / (2 * np.pi * self.KDE_bandwidth**2)

    def fit_transcripts(
        self,
        /,
        min_transcripts: float = 10,
        *,
        genes: Iterable[str] | None = None,
        fit_umap: bool = True,
    ):
        """
        Fits a spatial transcripts dataset using the SSAM algorithm.

        Parameters
        ----------
        min_transcripts : float
            Minimum expression for a local maximum to be considered. Expressed in terms
            of transcripts.
        genes : collections.abc.Iterable[str] | None
            A list of genes to utilize in the model. `None` uses all genes.
            Local maxima are always detected based on all genes.
        fit_umap : bool
            Whether to fit the UMAP to the data.
        """

        if genes is None:
            self.genes = sorted(self.transcripts["gene"].unique())
        else:
            self.genes = list(genes)

        local_maxima = _sample_expression(
            self.transcripts,
            min_expression=self._expression_threshold(min_transcripts),
            kde_bandwidth=self.KDE_bandwidth,
            genes=self.genes,
            n_workers=self.n_workers,
            min_pixel_distance=self.min_distance,
            patch_length=self.patch_length,
            dtype=self.dtype,
        )

        self.fit_pseudocells(local_maxima, fit_umap=fit_umap)

    def fit_pseudocells(self, pseudocells: AnnData, *, fit_umap: bool = True):
        """
        Fits the expression of pseudocells.

        Parameters
        ----------
        pseudocells : anndata.AnnData
            Gene expression to use for fitting.
        fit_umap : bool
            Whether to fit the UMAP to the data.
        """

        self.pseudocells = pseudocells
        X = pseudocells[:, self.genes].X
        self.pca.fit(X)

        if fit_umap:
            factors = self.pca.transform(X)

            print(f"Modeling {factors.shape[1]} pseudo-celltype clusters;")

            self.pseudocells.obsm["2D_UMAP"] = self.umap_2d.fit_transform(factors)

            embedding_color = self.umap_rgb.fit_transform(
                factors / norm(factors, axis=1, keepdims=True)
            )
            embedding_color = _fill_color_axes(embedding_color, self.pca_rgb, fit=True)

            self._colors_min_max = (
                embedding_color.min(axis=0),
                embedding_color.max(axis=0),
            )

            self.pseudocells.obsm["RGB"] = _minmax_scaling(embedding_color)

    @staticmethod
    def _determine_celltype(samples: AnnData, signatures: pl.DataFrame) -> np.ndarray:
        X = samples[:, signatures.columns].X
        signature_mtx = signatures.to_numpy()
        # TODO: this is quite inefficient?
        # it calculates all pairwise correlation coefficients even if not needed.
        correlations = np.array(
            [np.corrcoef(X[i, :], signature_mtx)[0, 1:] for i in range(X.shape[0])]
        )
        return np.argmax(correlations, axis=-1)

    @staticmethod
    def _coordinate_center(
        embedding: np.ndarray, assignments: np.ndarray, n: int
    ) -> np.ndarray:
        # determine the center of gravity of each assignment (celltype or gene)
        # in the embedding
        return np.array(
            [
                (
                    np.median(embedding[assignments == i, :], axis=0)
                    if (assignments == i).sum() > 0
                    else (np.nan, np.nan)
                )
                for i in range(n)
            ]
        )

    @overload
    def fit_signatures(self, signatures: pl.DataFrame, key: str): ...
    @overload
    def fit_signatures(self, signatures: pd.DataFrame, key: None = None): ...

    def fit_signatures(
        self, signatures: pl.DataFrame | pd.DataFrame, key: None | str = None
    ):
        """
        Fits a signature matrix.

        Parameters
        ----------
        signatures : polars.DataFrame | pandas.DataFrame
            A matrix of celltypes x gene signatures to use to annotate the UMAP.
        key : str | None
            Name of the column with name of the signature.
            Only used if `signatures` is a :py:class:`polars.DataFrame`,
            for :py:class:`pandas.DataFrame` the names are expected as index.
        """

        if self.pseudocells is None:
            raise Exception("`fit_pseudocells` must be run before fitting signatures")
        if "2D_UMAP" not in self.pseudocells.obsm:
            raise Exception("Signature can only be fitted if a UMAP has been fit")

        if isinstance(signatures, pl.DataFrame):
            assert key is not None
            # ensure annotation as first column
            signatures = signatures.select(key, pl.exclude(key))
        else:
            if signatures.index.name is None:
                key = "index"
            else:
                key = str(signatures.index.name)
            signatures = pl.from_pandas(signatures.reset_index())

        self.signatures = signatures

        self.celltype_assignments = self._determine_celltype(
            self.pseudocells, signatures.drop(key)
        )

        # determine the center of gravity of each celltype in the embedding:
        self.celltype_centers = self._coordinate_center(
            self.pseudocells.obsm["2D_UMAP"],
            self.celltype_assignments,
            signatures.shape[0],
        )

    def compute_VSI(self, *, min_transcripts: float = 2):
        """
        Calculate the vertical signal integrity (VSI).

        Parameters
        ----------
        min_transcripts : float | None, optional
            Minimum expression value to consider in calculation.
            Defaults to the 110% of the maximum expression profile of two molecules in the KDE.
        """

        min_expression = self._expression_threshold(min_transcripts)

        padding = int(ceil(_TRUNCATE * self.KDE_bandwidth))

        gene2idx = {gene: i for i, gene in enumerate(self.genes)}

        signal = kde_2d_discrete(
            self.transcripts["x_pixel"].to_numpy(),
            self.transcripts["y_pixel"].to_numpy(),
            bandwidth=self.KDE_bandwidth,
            dtype=self.dtype,
        )
        shape = signal.shape

        cosine_similarity = np.zeros_like(signal)

        with ThreadPoolExecutor(max_workers=self.n_workers) as executor:
            for patch_df, padded, unpadded in tqdm.tqdm(
                _patches(
                    self.transcripts[["gene", "x_pixel", "y_pixel", "z", "z_center"]],
                    self.patch_length,
                    padding,
                    size=shape,
                    coordinates=("x_pixel", "y_pixel"),
                ),
                total=n_patches(self.patch_length, shape),
            ):
                assert isinstance(patch_df, pl.DataFrame)

                if len(patch_df) == 0:
                    continue

                # remove padding
                left_pad = unpadded[0].start - padded[0].start
                bottom_pad = unpadded[1].start - padded[1].start
                x_size = unpadded[0].stop - unpadded[0].start
                y_size = unpadded[1].stop - unpadded[1].start
                remove_pad = (
                    slice(left_pad, left_pad + x_size),
                    slice(bottom_pad, bottom_pad + y_size),
                )

                patch_signal: KDE_2D = signal[padded]

                not_padding = np.zeros(patch_signal.shape, dtype=bool)
                not_padding[remove_pad] = True

                patch_mask = (patch_signal > min_expression) & not_padding
                n_pixels = patch_mask.sum()

                if n_pixels == 0:
                    continue

                patch_df = patch_df.filter(
                    pl.col("gene").cast(pl.String).is_in(gene2idx)
                )
                gene_queue: SimpleQueue[tuple[int, pl.DataFrame]] = SimpleQueue()
                for (gene, *_), df in patch_df.group_by("gene"):
                    if gene in gene2idx:
                        gene_queue.put((gene2idx[gene], df.drop("gene")))

                embedding_top, embedding_bottom = reduce(
                    lambda x, y: (x[0] + y[0], x[1] + y[1]),  # type: ignore
                    map(
                        lambda x: x.result(),
                        as_completed(
                            executor.submit(
                                _calculate_embedding,
                                gene_queue,
                                patch_mask,
                                self.pca.components_,
                                bandwidth=self.KDE_bandwidth,
                                dtype=self.dtype,
                            )
                            for _ in range(self.n_workers)
                        ),
                    ),
                )

                if (isinstance(embedding_top, int) and embedding_top == 0) or (
                    isinstance(embedding_bottom, int) and embedding_bottom == 0
                ):
                    continue
                assert isinstance(embedding_top, np.ndarray)
                assert isinstance(embedding_bottom, np.ndarray)
                patch_cosine_similarity = np.zeros_like(patch_signal)
                patch_cosine_similarity[patch_mask] = _cosine_similarity(
                    embedding_top, embedding_bottom
                )
                cosine_similarity[unpadded] = patch_cosine_similarity[remove_pad]

        self.signal_map = signal.T
        self.integrity_map = cosine_similarity.T

    def analyse(
        self,
        gridsize: float = 1,
        min_transcripts: float = 10,
        genes: Iterable[str] | None = None,
        fit_umap: bool = True,
    ):
        """
        Run main ovrlpy analysis.

        Parameters
        ----------
        gridsize : float, optional
            The size of the pixel grid.
        min_transcripts : float
            Minimum expression for a local maximum to be considered. Expressed in terms
            of transcripts.
        genes : collections.abc.Iterable[str] | None
            A list of genes to utilize in the model. `None` uses all genes.
            Local maxima are always detected based on all genes.
        fit_umap : bool
            Whether to fit the UMAP to the data.
        """

        print("Running vertical adjustment")
        self.process_coordinates(gridsize=gridsize, n_iter=20)

        print("Creating gene expression embeddings for visualization")
        self.fit_transcripts(min_transcripts, genes=genes, fit_umap=fit_umap)

        print("Creating signal integrity map")
        self.compute_VSI()

    def detect_doublets(
        self,
        min_distance: int = 10,
        min_integrity: float = 0.7,
        min_signal: float = 3,
        integrity_sigma: float | None = None,
    ) -> pl.DataFrame:
        """
        This function is used to find individual low peaks of signal integrity in the tissue
        map as an indicator of single occurrences overlapping cells.

        Parameters
        ----------
        min_distance : int, optional
            Minimum distance between reported peaks
        min_integrity : float, optional
            Threshold of signal integrity value. A peak with an
            `signal_integrity < min_integrity` is not considered.
        min_signal : float, optional
            Minimum signal value for a peak to be considered
        integrity_sigma : float, optional
            Optional sigma value for gaussian filtering of the integrity map,
            which leads to the detection of overlap regions with larger spatial extent.

        Returns
        -------
        polars.DataFrame
        """

        if not hasattr(self, "signal_map") or not hasattr(self, "integrity_map"):
            raise Exception("Run `compute_VSI` before detecting doublets")

        if integrity_sigma is not None:
            integrity_map = gaussian_filter(self.integrity_map, integrity_sigma)
        else:
            integrity_map = self.integrity_map

        dist_x, dist_y, dist_t = _determine_localmax_and_sample(
            (1 - integrity_map) * (self.signal_map > min_signal),
            min_distance=min_distance,
            min_value=1 - min_integrity,
        )

        doublets = pl.DataFrame(
            {
                "x": dist_y,
                "y": dist_x,
                "integrity": 1 - dist_t,
                "signal": self.signal_map[dist_x, dist_y],
            }
        ).sort("integrity")

        return doublets

    def subset_transcripts(
        self, x: float, y: float, *, window_size: int = 30
    ) -> pl.DataFrame:
        """
        Subset the transcript dataframe spatially based on given x, y coordinates and window
        size.

        Parameters
        ----------
        x : float
            x-coordinate to center the region
        y : float
            y-coordinate to center the region
        window_size : int, optional
            The window size of the region. Molecules within this window around (x, y)
            are returned as a new DataFrame.

        Returns
        -------
        polars.DataFrame
        """
        return self.transcripts.filter(
            pl.col("x").is_between(x - window_size, x + window_size)
            & pl.col("y").is_between(y - window_size, y + window_size)
        ).clone()

    def transform_transcripts(
        self,
        transcripts: pl.DataFrame | pd.DataFrame,
        *,
        gene_key: str = "gene",
        coordinate_keys: Sequence[str] = ["x", "y", "z"],
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Transforms the coordinate dataframe to the 2D and 3D embedding space.

        Parameters
        ----------
        transcripts : polars.DataFrame | pandas.DataFrame
            Dataframe of transcript coordinates to transform.
        gene_key : str
            Name of the gene column.
        coordinate_keys : collections.abc.Sequence[str]
            Names of the coordinate columns.

        Returns
        -------
        embedding : numpy.ndarray
            2D UMAP embedding
        rgb : numpy.ndarray
            3D RGB UMAP embedding
        """

        neighbors = _gaussian_weighted_neighbors(
            transcripts[list(coordinate_keys)],
            self.KDE_bandwidth * _TRUNCATE,
            self.KDE_bandwidth,
            n_workers=self.n_workers,
        )

        expression = _weighted_average_expression(
            pl.Series(transcripts[gene_key]), neighbors, self.genes
        )
        return self.transform_pseudocells(expression)

    def transform_pseudocells(
        self, pseudocells: pl.DataFrame | pd.DataFrame
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Transforms a matrix of gene expression to the 2D and 3D embedding space.

        Parameters
        ----------
        pseudocells : polars.DataFrame | pandas.DataFrame
            A cell x gene matrix of gene expression

        Returns
        -------
        embedding : numpy.ndarray
            2D UMAP embedding
        rgb : numpy.ndarray
            3D RGB UMAP embedding
        """

        embedding, embedding_color = _transform_embeddings(
            pseudocells.to_numpy(),
            self.pca,
            umap_2d=self.umap_2d,
            umap_rgb=self.umap_rgb,
        )
        embedding_color = _fill_color_axes(embedding_color, self.pca_rgb)
        color_min, color_max = self._colors_min_max
        embedding_color = (embedding_color - color_min) / (color_max - color_min)
        embedding_color = np.clip(embedding_color, 0, 1)

        return embedding, embedding_color

    def pseudocell_integrity(self) -> pl.DataFrame:
        """
        Returns a DataFrame containing the gene-count matrix of the fitted
        tissue's determined pseudo-cells.

        Returns
        -------
        polars.DataFrame
        """
        assert self.pseudocells is not None
        pseudocells = pl.DataFrame(
            self.pseudocells.obsm["spatial"][:, :2], schema=["x", "y"]
        )

        if self.signal_map is not None:
            pseudocells = pseudocells.with_columns(
                signal=self.signal_map[pseudocells["y"], pseudocells["x"]]
            )

        if self.integrity_map is not None:
            pseudocells = pseudocells.with_columns(
                integrity=self.integrity_map[pseudocells["y"], pseudocells["x"]]
            )

        return pseudocells



from importlib.metadata import PackageNotFoundError, version

from . import io
from ._ovrlp import Ovrlp
from ._plotting import (
    SCALEBAR_PARAMS,
    plot_pseudocells,
    plot_region_of_interest,
    plot_signal_integrity,
    plot_tissue,
    plot_umap,
)
from ._subslicing import process_coordinates
from ._utils import UMAP_2D_PARAMS, UMAP_RGB_PARAMS

try:
    __version__ = version("ovrlpy")
except PackageNotFoundError:
    __version__ = "unknown version"

del PackageNotFoundError, version


__all__ = [
    "io",
    "Ovrlp",
    "plot_pseudocells",
    "plot_region_of_interest",
    "plot_signal_integrity",
    "plot_tissue",
    "plot_umap",
    "process_coordinates",
    "SCALEBAR_PARAMS",
    "UMAP_2D_PARAMS",
    "UMAP_RGB_PARAMS",
]



import warnings
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor, as_completed
from math import floor
from typing import TypeAlias, TypeVar

import numpy as np
import pandas as pd
import polars as pl
import tqdm
from anndata import AnnData, ImplicitModificationWarning
from numpy.typing import DTypeLike
from scipy.ndimage import gaussian_filter
from scipy.sparse import coo_array
from skimage.feature import peak_local_max

from ._patching import _patches, n_patches

_TRUNCATE = 4

Shape1D = tuple[int]
Array1D = np.ndarray[Shape1D, np.dtype]
Array1D_Int = np.ndarray[Shape1D, np.dtype[np.integer]]

Coordinate: TypeAlias = Array1D | pl.Series

KDE_2D = np.ndarray[tuple[int, int], np.dtype[np.floating]]
KDE_ND = np.ndarray[tuple[int, ...], np.dtype[np.floating]]


def kde_2d_discrete(
    x: Array1D_Int,
    y: Array1D_Int,
    bandwidth: float,
    size: tuple[int, int] | None = None,
    dtype: DTypeLike = np.float32,
    **kwargs,
) -> KDE_2D:
    """
    Calculate the 2D KDE using discrete (i.e. integer) coordinates.
    """
    n = x.shape[0]
    if x.shape[0] != y.shape[0]:
        raise ValueError("All coordinates must have the same number of rows")

    if n == 0:
        if size is None:
            raise ValueError("If no coordinates are provided, size must be provided.")
        else:
            return np.zeros(size, dtype=dtype)

    if size is None:
        size = (x.max() + 1, y.max() + 1)

    min_x = x.min()
    min_y = y.min()
    if min_x != 0:
        x = x - min_x
    if min_y != 0:
        y = y - min_y

    counts = coo_array((np.ones(n, dtype=np.uint32), (x, y))).toarray()
    kde = _kde(counts, bandwidth, dtype=dtype, **kwargs)

    if kde.shape != size:
        output = np.zeros(size, dtype=dtype)
        output[min_x : kde.shape[0] + min_x, min_y : kde.shape[1] + min_y] = kde
        return output
    else:
        return kde


def kde_nd(
    *coordinates: Coordinate,
    bandwidth: float,
    size: tuple[int, ...] | None = None,
    dtype: DTypeLike = np.float32,
    **kwargs,
) -> KDE_ND:
    """
    Calculate the KDE using the (continuous) coordinates.
    """
    assert len(coordinates) >= 1

    mins = tuple(int(floor(c.min())) for c in coordinates)
    maxs = tuple(int(floor(c.max() + 1)) for c in coordinates)
    assert all(min >= 0 for min in mins)

    n = coordinates[0].shape[0]
    if not all(x.shape[0] == n for x in coordinates[1:]):
        raise ValueError("All coordinates must have the same number of rows")

    if n == 0:
        if size is None:
            raise ValueError("If no coordinates are provided, size must be provided.")
        else:
            return np.zeros(size, dtype=dtype)

    if size is None:
        size = maxs

    dim_bins = [np.arange(min, max + 1) for min, max in zip(mins, maxs)]
    counts, bins = np.histogramdd(coordinates, bins=dim_bins)
    kde = _kde(counts, bandwidth, dtype=dtype, **kwargs)

    if kde.shape != size:
        output = np.zeros(size, dtype=dtype)
        output[tuple(slice(b[0], b[-1]) for b in bins)] = kde
        return output
    else:
        return kde


def _kde(
    x: np.ndarray,
    bandwidth: float,
    truncate: float = _TRUNCATE,
    dtype: DTypeLike = np.float32,
) -> np.ndarray:
    kde = gaussian_filter(
        x, sigma=bandwidth, truncate=truncate, mode="constant", output=dtype
    )
    return kde


def find_local_maxima(
    x: np.ndarray, min_pixel_distance: int = 5, min_expression: float = 2
):
    local_maxima = peak_local_max(
        x,
        min_distance=min_pixel_distance,
        threshold_abs=min_expression,
        exclude_border=False,
    )

    return local_maxima


T = TypeVar("T")


def kde_and_sample(
    *coordinates: Coordinate, sampling_coordinates: np.ndarray, gene: T, **kwargs
) -> tuple[T, np.ndarray]:
    """
    Create a kde of the data and sample at 'sampling_coordinates'.
    """

    sampling_coordinates = np.rint(sampling_coordinates).astype(int)
    n_dims = sampling_coordinates.shape[1]

    kde = kde_nd(*coordinates, **kwargs)

    return gene, kde[tuple(sampling_coordinates[:, i] for i in range(n_dims))]


def _sample_expression(
    transcripts: pl.DataFrame,
    kde_bandwidth: float = 2.5,
    min_expression: float = 2,
    min_pixel_distance: float = 5,
    genes: list[str] | None = None,
    coord_columns: Iterable[str] = ["x", "y", "z"],
    gene_column: str = "gene",
    n_workers: int = 8,
    patch_length: int = 500,
    dtype: DTypeLike = np.float32,
) -> AnnData:
    """
    Sample expression from a transcripts dataframe.

    Parameters
    ----------
    transcripts : pandas.DataFrame
        The input transcripts dataframe.
    kde_bandwidth : float
        Bandwidth for kernel density estimation.
    minimum_expression : int
        Minimum expression value for local maxima determination.
    min_pixel_distance : int
        Minimum pixel distance for local maxima determination.
    genes : list[str] | None
        Which genes to use in sampling of the local maxima. Detection of local maxima
        will be done on all genes.
    coord_columns : Iterable[str], optional
        Name of the coordinate columns in the coordinate dataframe.
    gene_column : str, optional
        Name of the gene column in the coordinate dataframe.
    n_workers : int, optional
        Number of parallel workers for sampling.
    patch_length : int
        Size of the length in each dimension when calculating signal integrity in patches.
        Smaller values will use less memory, but may take longer to compute.
    dtype : numpy.typing.DTypeLike
        Datatype for the KDE.

    Returns
    -------
    anndata.AnnData
    """

    coord_columns = list(coord_columns)
    assert len(coord_columns) == 3 or len(coord_columns) == 2

    # lower resolution instead of increasing bandwidth!
    transcripts = (
        transcripts.lazy()
        .select(pl.col(coord_columns) / kde_bandwidth, gene_column)
        .collect(engine="streaming")
    )

    print("determining pseudocells")

    # perform a global KDE to determine local maxima:
    kde = kde_nd(*(transcripts[c] for c in coord_columns), bandwidth=1, dtype=dtype)

    min_dist = 1 + int(min_pixel_distance / kde_bandwidth)
    local_maximum_coordinates = find_local_maxima(
        kde, min_pixel_distance=min_dist, min_expression=min_expression
    )

    print("found", len(local_maximum_coordinates), "pseudocells")

    size = kde.shape
    del kde

    if genes is not None:
        transcripts = transcripts.filter(pl.col("gene").cast(pl.String).is_in(genes))
    gene_list = sorted(transcripts[gene_column].unique())

    # truncate * bandwidth -> _TRUNCATE * 1
    padding = _TRUNCATE

    print("sampling expression:")
    patches = []
    coords = []
    with ThreadPoolExecutor(max_workers=n_workers) as executor:
        for patch_df, padded, unpadded in tqdm.tqdm(
            _patches(transcripts, patch_length, padding, size=size),
            total=n_patches(patch_length, size),
        ):
            assert isinstance(patch_df, pl.DataFrame)
            patch_maxima = local_maximum_coordinates[
                (local_maximum_coordinates[:, 0] >= unpadded[0].start)
                & (local_maximum_coordinates[:, 0] < unpadded[0].stop)
                & (local_maximum_coordinates[:, 1] >= unpadded[1].start)
                & (local_maximum_coordinates[:, 1] < unpadded[1].stop),
                :,
            ]
            coords.append(patch_maxima)

            # we need to shift the maximum coordinates so they are in the correct
            # relative position of the patch
            maxima = patch_maxima.copy()
            maxima[:, 0] -= padded[0].start
            maxima[:, 1] -= padded[1].start

            # patch_size is 2D, make 3D if KDE is calculated as 3D
            patch_size = (
                padded[0].stop - padded[0].start,
                padded[1].stop - padded[1].start,
                *size[2:],
            )

            futures = set(
                executor.submit(
                    kde_and_sample,
                    *(df[c] for c in coord_columns),
                    sampling_coordinates=maxima,
                    gene=gene[0],
                    size=patch_size,
                    bandwidth=1,
                    dtype=dtype,
                )
                for gene, df in patch_df.group_by(gene_column)
            )

            # TODO: improve
            patches.append(
                pd.DataFrame(dict(f.result() for f in as_completed(futures)))
            )
            del futures
        del transcripts, local_maximum_coordinates

    # TODO: sparse?
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ImplicitModificationWarning)
        adata = AnnData(pd.concat(patches, ignore_index=True)[gene_list].fillna(0))
    adata.obsm["spatial"] = np.rint(np.vstack(coords) * kde_bandwidth).astype(np.int32)
    return adata



import os
from collections.abc import Collection, Mapping
from pathlib import Path

import polars as pl


def _filter_genes(df: pl.DataFrame, remove_features: Collection[str]) -> pl.DataFrame:
    if len(remove_features) > 0:
        remove_pattern = "|".join(remove_features)
        df = (
            df.lazy()
            .with_columns(pl.col("gene").cast(pl.String))
            .filter(~pl.col("gene").str.contains(remove_pattern))
            .with_columns(pl.col("gene").cast(pl.Categorical))
            .collect()
        )
    return df


# 10x Xenium
_XENIUM_COLUMNS = {
    "feature_name": "gene",
    "x_location": "x",
    "y_location": "y",
    "z_location": "z",
}

XENIUM_CTRLS = [
    "^BLANK",
    "^DeprecatedCodeword",
    "^Intergenic",
    "^NegControl",
    "^UnassignedCodeword",
]
"""Patterns for Xenium controls"""


def read_Xenium(
    filepath: str | os.PathLike,
    *,
    min_qv: float | None = None,
    remove_features: Collection[str] = XENIUM_CTRLS,
    additional_columns: Collection[str] = [],
    n_threads: int | None = None,
) -> pl.DataFrame:
    """
    Read a Xenium transcripts file.

    Parameters
    ----------
    filepath : os.PathLike or str
        Path to the Xenium transcripts file. Both, .csv.gz and .parquet files, are supported.
    min_qv : float | None, optional
        Minimum Phred-scaled quality value (Q-Score) of a transcript to be included.
        If `None` no filtering is performed.
    remove_features : collections.abc.Collection[str], optional
        List of regex patterns to filter the 'feature_name' column,
        :py:attr:`ovrlpy.io.XENIUM_CTRLS` by default.
    additional_columns : collections.abc.Collection[str], optional
        Additional columns to load from the transcripts file.
    n_threads : int | None, optional
        Number of threads used for parsing the input file.
        If None, will default to number of available CPUs.

    Returns
    -------
    polars.DataFrame
    """
    filepath = Path(filepath)
    columns = list(set(_XENIUM_COLUMNS.keys()) | set(additional_columns))

    if filepath.suffix == ".parquet":
        transcripts = pl.scan_parquet(filepath)

        # 'is_gene' column only exists for Xenium v3 which only has .parquet
        if "is_gene" in transcripts.collect_schema().names():
            transcripts = transcripts.filter(pl.col("is_gene"))

        if min_qv is not None:
            transcripts = transcripts.filter(pl.col("qv") >= min_qv)

        with pl.StringCache():
            transcripts = (
                transcripts.select(columns)
                .with_columns(
                    pl.col("feature_name").cast(pl.String).cast(pl.Categorical)
                )
                .collect()
            )

    else:
        if min_qv is not None and "qv" not in additional_columns:
            columns.append("qv")
        transcripts = pl.read_csv(
            filepath,
            columns=columns,
            schema_overrides={"feature_name": pl.Categorical},
            n_threads=n_threads,
        )

        if min_qv is not None:
            transcripts = transcripts.filter(pl.col("qv") >= min_qv)
            if "qv" not in additional_columns:
                transcripts = transcripts.drop("qv")

    transcripts = transcripts.rename(_XENIUM_COLUMNS)
    transcripts = _filter_genes(transcripts, remove_features)

    return transcripts


# Vizgen MERSCOPE
_MERSCOPE_COLUMNS = {"gene": "gene", "global_x": "x", "global_y": "y", "global_z": "z"}

MERSCOPE_CTRLS = ["^Blank"]
"""Patterns for Vizgen controls"""


def read_MERSCOPE(
    filepath: str | os.PathLike,
    z_scale: float = 1.5,
    *,
    remove_genes: Collection[str] = MERSCOPE_CTRLS,
    additional_columns: Collection[str] = [],
    n_threads: int | None = None,
) -> pl.DataFrame:
    """
    Read a Vizgen transcripts file.

    Parameters
    ----------
    filepath : os.PathLike or str
        Path to the Vizgen transcripts file. Both, .csv(.gz) and .parquet files, are supported.
    z_scale : float
        Factor to scale z-plane index to um, i.e. distance between z-planes.
    remove_genes : collections.abc.Collection[str], optional
        List of regex patterns to filter the 'gene' column,
        :py:attr:`ovrlpy.io.MERSCOPE_CTRLS` by default.
    additional_columns : collections.abc.Collection[str], optional
        Additional columns to load from the transcripts file.
    n_threads : int | None, optional
        Number of threads used for parsing the input file.
        If None, will default to number of available CPUs.

    Returns
    -------
    polars.DataFrame
    """
    filepath = Path(filepath)
    columns = list(set(_MERSCOPE_COLUMNS.keys()) | set(additional_columns))

    if filepath.suffixes[-2:] == [".csv", ".gz"]:
        transcripts = pl.read_csv(
            filepath,
            columns=columns,
            schema_overrides={"gene": pl.Categorical},
            n_threads=n_threads,
        )

    else:
        if filepath.suffix == ".parquet":
            transcripts = pl.scan_parquet(filepath)
        elif filepath.suffix == ".csv":
            transcripts = pl.scan_csv(filepath)
        else:
            raise ValueError(
                "Unsupported file format; must be one of .csv(.gz) or .parquet"
            )

        with pl.StringCache():
            transcripts = (
                transcripts.select(columns)
                .with_columns(pl.col("gene").cast(pl.String).cast(pl.Categorical))
                .collect()
            )

    transcripts = transcripts.rename(_MERSCOPE_COLUMNS)
    transcripts = _filter_genes(transcripts, remove_genes)

    # convert plane to um
    transcripts = transcripts.with_columns(pl.col("z") * z_scale)

    return transcripts


# Nanostring CosMx
_COSMX_COLUMNS = {"target": "gene", "x_global_px": "x", "y_global_px": "y", "z": "z"}

COSMX_CTRLS = ["^NegPrb"]
"""Patterns for CosMx controls"""


def read_CosMx(
    filepath: str | os.PathLike,
    scale: Mapping[str, float] = {"xy": 0.12028, "z": 0.8},
    *,
    remove_targets: Collection[str] = COSMX_CTRLS,
    additional_columns: Collection[str] = [],
    n_threads: int | None = None,
) -> pl.DataFrame:
    """
    Read a Nanostring CosMx transcripts file.

    Parameters
    ----------
    filepath : os.PathLike or str
        Path to the CosMx transcripts file.
    scale : collections.abc.Mapping[str, float]
        The factors for scaling the coordinates from pixel space to um.
    remove_targets : collections.abc.Collection[str], optional
        List of regex patterns to filter the 'target' column,
        :py:attr:`ovrlpy.io.COSMX_CTRLS` by default.
    additional_columns : collections.abc.Collection[str], optional
        Additional columns to load from the transcripts file.
    n_threads : int | None, optional
        Number of threads used for parsing the input file.
        If None, will default to number of available CPUs.

    Returns
    -------
    polars.DataFrame
    """

    transcripts = pl.read_csv(
        Path(filepath),
        columns=list(set(_COSMX_COLUMNS.keys()) | set(additional_columns)),
        schema_overrides={"target": pl.Categorical},
        n_threads=n_threads,
    ).rename(_COSMX_COLUMNS)

    transcripts = _filter_genes(transcripts, remove_targets)

    # convert pixel to um
    transcripts = transcripts.with_columns(
        pl.col(["x", "y"]) * scale["xy"], pl.col("z") * scale["z"]
    )

    return transcripts



from collections.abc import Sequence
from math import floor

import numpy as np
import polars as pl


def ceildiv(a: int, b: int) -> int:
    return -(a // -b)


def n_patches(length: int, size: Sequence[int]) -> int:
    return ceildiv(size[0], length) * ceildiv(size[1], length)


def _patches(
    df: pl.DataFrame,
    length: int,
    padding: int,
    *,
    size: None | Sequence[int] = None,
    coordinates: tuple[str, str] = ("x", "y"),
):
    x_key, y_key = coordinates

    if size is None:
        size = (int(floor(df[x_key].max() + 1)), int(floor(df[y_key].max() + 1)))

    # ensure that patch_length is an upper-bound for the actual size
    patch_count_x = ceildiv(size[0], length)
    patch_count_y = ceildiv(size[1], length)

    x_patches = np.linspace(0, size[0], patch_count_x + 1, dtype=int)
    y_patches = np.linspace(0, size[1], patch_count_y + 1, dtype=int)

    for i in range(len(x_patches) - 1):
        for j in range(len(y_patches) - 1):
            # coordinates with padding
            x_ = max(0, x_patches[i] - padding)
            y_ = max(0, y_patches[j] - padding)
            _x = min(size[0], x_patches[i + 1] + padding)
            _y = min(size[1], y_patches[j + 1] + padding)

            padded_range = (slice(x_, _x), slice(y_, _y))

            patch = df.filter(
                pl.col(x_key).is_between(x_, _x, closed="left")
                & pl.col(y_key).is_between(y_, _y, closed="left")
            ).with_columns(pl.col(x_key) - x_, pl.col(y_key) - y_)

            unpadded_range = (
                slice(x_patches[i], x_patches[i + 1]),
                slice(y_patches[j], y_patches[j + 1]),
            )

            yield patch, padded_range, unpadded_range



from collections.abc import Iterable, Mapping
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import polars as pl
from matplotlib.axes import Axes
from matplotlib.colors import Colormap, LinearSegmentedColormap
from matplotlib.figure import Figure
from matplotlib.image import AxesImage
from matplotlib.patches import Rectangle
from matplotlib.text import Text
from matplotlib_scalebar.scalebar import ScaleBar

from ._ovrlp import Ovrlp

SCALEBAR_PARAMS: dict[str, Any] = {"dx": 1, "units": "um"}
"""Default scalebar parameters"""

BIH_CMAP = LinearSegmentedColormap.from_list(
    "BIH",
    [
        "#430541",
        "mediumvioletred",
        "violet",
        "powderblue",
        "powderblue",
        "white",
        "white",
    ][::-1],
)

VSI = "vertical signal integrity"
_SIGNAL_THRESHOLD = 2


def _plot_scalebar(ax: Axes, dx: float = 1, units="um", **kwargs):
    ax.add_artist(ScaleBar(dx, units=units, **kwargs))


def _untangle_text(
    text_artists: list[Text], ax: Axes | None = None, max_iterations: int = 10_000
):
    if ax is None:
        ax = plt.gca()
    inv = ax.transData.inverted()

    artist_coords = np.array(
        [text_artist.get_position() for text_artist in text_artists]
    )
    artist_coords = artist_coords + np.random.normal(0, 0.001, artist_coords.shape)
    artist_bbox = [text_artist.get_window_extent() for text_artist in text_artists]
    artist_extents = np.array(
        [inv.transform(bbox.get_points()) for bbox in artist_bbox]
    )
    artist_extents = artist_extents[:, 1] - artist_extents[:, 0]

    for i in range(max_iterations):
        relative_positions_x = (
            artist_coords[:, 0][:, None] - artist_coords[:, 0][None, :]
        )
        relative_positions_y = (
            artist_coords[:, 1][:, None] - artist_coords[:, 1][None, :]
        )

        relative_positions_x /= (
            0.1 + (artist_extents[:, 0][:, None] + artist_extents[:, 0][None, :]) / 2
        )
        relative_positions_y /= (
            0.1 + (artist_extents[:, 1][:, None] + artist_extents[:, 1][None, :]) / 2
        )

        # distances = np.sqrt(relative_positions_x**2+relative_positions_y**2)
        distances = np.abs(relative_positions_x) + np.abs(relative_positions_y)

        gaussian_repulsion = 1 * np.exp(-distances / 0.5)

        velocities_x = np.zeros_like(relative_positions_x)
        velocities_y = np.zeros_like(relative_positions_y)

        velocities_x[distances > 0] = (
            gaussian_repulsion[distances > 0]
            * relative_positions_x[distances > 0]
            / distances[distances > 0]
        )
        velocities_y[distances > 0] = (
            gaussian_repulsion[distances > 0]
            * relative_positions_y[distances > 0]
            / distances[distances > 0]
        )

        velocities_x[np.eye(velocities_x.shape[0], dtype=bool)] = 0
        velocities_y[np.eye(velocities_y.shape[0], dtype=bool)] = 0

        delta = np.stack([velocities_x, velocities_y], axis=1).mean(-1)
        # # delta = delta.clip(-0.1,0.1)
        artist_coords = artist_coords + delta * 0.1
        # artist_coords  = artist_coords*0.9 + initial_artist_coords*0.1

    for i, text_artist in enumerate(text_artists):
        text_artist.set_position(artist_coords[i, :])


def plot_umap(ovrlp: Ovrlp, *, annotate: bool = True, ax: Axes | None = None, **kwargs):
    """
    Plots the UMAP embedding.

    Parameters
    ----------
    annotate : bool
        Whether to add cell-type annotation if it was calculated using `Ovrlp.fit_signatures`.
    ax : matplotlib.axes.Axes | None
        Axis object to plot on.
    kwargs
        Keyword arguments for :py:func:`matplotlib.pyplot.scatter`.
    """
    if ax is None:
        _, ax = plt.subplots()
    annotation = (
        ovrlp.signatures[:, 0] if annotate and hasattr(ovrlp, "signatures") else None
    )
    assert ("2D_UMAP" in ovrlp.pseudocells.obsm) and ("RGB" in ovrlp.pseudocells.obsm)
    ct_center = ovrlp.celltype_centers if hasattr(ovrlp, "celltype_centers") else None
    _plot_embeddings(
        ax,
        ovrlp.pseudocells.obsm["2D_UMAP"],
        ovrlp.pseudocells.obsm["RGB"],
        ct_center,
        annotation,
        **kwargs,
    )


# define a function that plots the embeddings, with celltype centers rendered as plt.texts on top:
def _plot_embeddings(
    ax: Axes,
    embedding: np.ndarray,
    color: np.ndarray,
    celltype_centers: np.ndarray | None = None,
    celltypes: Iterable[str] | None = None,
    scatter_kwargs: dict[str, Any] = {"alpha": 0.1, "marker": "."},
):
    ax.axis("off")

    alpha = scatter_kwargs.pop("alpha", 0.1)
    marker = scatter_kwargs.pop("marker", ".")

    ax.scatter(
        embedding[:, 0],
        embedding[:, 1],
        c=color,
        alpha=alpha,
        marker=marker,
        **scatter_kwargs,
    )
    ax.set(aspect="equal")

    if celltypes is not None and celltype_centers is not None:
        text_artists: list[Text] = []
        for i, celltype in enumerate(celltypes):
            if not np.isnan(celltype_centers[i, 0:2]).any():
                x, y = celltype_centers[i, 0:2]
                text = ax.text(x, y, celltype, color="k")
                text_artists.append(text)

        _untangle_text(text_artists, ax)


def plot_tissue(
    ovrlp: Ovrlp,
    *,
    scalebar: dict | None = SCALEBAR_PARAMS,
    ax: Axes | None = None,
    **kwargs,
):
    """
    Plots the tissue colored by the UMAP embedding.

    Parameters
    ----------
    scalebar : dict[str, typing.Any] | None
        If `None` no scalebar will be plotted. Otherwise a dictionary with
        additional kwargs for ``matplotlib_scalebar.scalebar.ScaleBar``.
        By default :py:attr:`ovrlpy.SCALEBAR_PARAMS`
    ax : matplotlib.axes.Axes | None
        Axis object to plot on.
    kwargs
        Keyword arguments for the matplotlib's scatter plot function.
    """
    if ax is None:
        _, ax = plt.subplots()

    _plot_tissue_scatter(
        ax,
        ovrlp.pseudocells.obsm["spatial"][:, 0],
        ovrlp.pseudocells.obsm["spatial"][:, 1],
        ovrlp.pseudocells.obsm["RGB"],
        marker=".",
        **kwargs,
    )

    if scalebar is not None:
        _plot_scalebar(ax, **scalebar)


def _plot_tissue_scatter(ax: Axes, xs, ys, cs, *, title: str | None = None, **kwargs):
    ax.scatter(xs, ys, c=cs, **kwargs)
    ax.set_aspect("equal", adjustable="box")
    if title is not None:
        ax.set_title(title)


def _mark_roi_center(ax: Axes, x, y, roi):
    ax.scatter(x, y, c="k", marker="+", s=100)
    ax.set(xlim=roi[0], ylim=roi[1])


def plot_pseudocells(
    ovrlp: Ovrlp,
    *,
    umap_kwargs: Mapping[str, Any] = {"scatter_kwargs": {"s": 1}},
    tissue_kwargs: Mapping[str, Any] = {"s": 1},
    figsize: tuple[float, float] = (12, 6),
    **kwargs,
) -> Figure:
    """
    Plots the UMAP and tissue side-to-side.

    Parameters
    ----------
    ovrlp : ovrlpy.Ovrlp
    umap_kwargs
        Keyword arguments for :py:func:`ovrlpy.plot_umap`.
    tissue_kwargs
        Keyword arguments for :py:func:`ovrlpy.plot_tissue`.
    figsize : tuple[float, float]
        Size of the figure in inches.
    kwargs
        Other keyword arguments will be passed to :py:func:`matplotlib.pyplot.subplots`

    Returns
    -------
    matplotlib.figure.Figure
    """

    fig, axs = plt.subplots(figsize=figsize, ncols=2, **kwargs)

    plot_umap(ovrlp, ax=axs[0], **umap_kwargs)
    plot_tissue(ovrlp, ax=axs[1], **tissue_kwargs)

    return fig


def _plot_signal_integrity(
    ax: Axes,
    integrity: np.ndarray,
    signal: np.ndarray,
    threshold: float,
    *,
    cmap: Colormap = BIH_CMAP,
) -> AxesImage:
    # fade out for pixels with signal < threshold
    alpha = (signal / threshold).clip(0, 1) ** 2
    img = ax.imshow(integrity, cmap=cmap, alpha=alpha, vmin=0, vmax=1, origin="lower")
    return img


def plot_signal_integrity(
    ovrlp: Ovrlp,
    *,
    signal_threshold: float = _SIGNAL_THRESHOLD,
    cmap: str | Colormap = BIH_CMAP,
    histogram: bool = True,
    scalebar: dict | None = SCALEBAR_PARAMS,
    figsize: tuple[float, float] = (9, 6),
    **kwargs,
) -> Figure:
    """
    Plots the determined signal integrity of the tissue sample in a signal integrity map.

    Parameters
    ----------
    ovrlp : Ovrlp
    signal_threshold : float, optional
        Threshold below which the signal is faded out in the plot,
        to avoid displaying noisy areas with low predictive confidence.
        Pixels below the threshold are not counted in the histogram.
    cmap : str | matplotlib.colors.Colormap, optional
        Colormap for display.
    histogram : bool, optional
        Whether to plot a histogram of integrity values alongside the map.
    scalebar : dict[str, typing.Any] | None
        If `None` no scalebar will be plotted. Otherwise a dictionary with
        additional kwargs for ``matplotlib_scalebar.scalebar.ScaleBar``.
        By default :py:attr:`ovrlpy.SCALEBAR_PARAMS`
    figsize : tuple[float, float]
        Size of the figure in inches.
    kwargs
        Other keyword arguments are passed to :py:func:`matplotlib.pyplot.subplots`

    Returns
    -------
    matplotlib.figure.Figure
    """
    integrity = ovrlp.integrity_map
    signal = ovrlp.signal_map

    cmap = Colormap(cmap) if isinstance(cmap, str) else cmap

    with plt.style.context("dark_background"):
        if histogram:
            kwargs = {"width_ratios": [6, 1]} | kwargs
            fig, axs = plt.subplots(ncols=2, figsize=figsize, **kwargs)
            ax_im, ax_hist = axs
            assert isinstance(ax_hist, Axes)
        else:
            fig, ax_im = plt.subplots(figsize=figsize, **kwargs)

        assert isinstance(ax_im, Axes)

        img = _plot_signal_integrity(
            ax_im, integrity, signal, signal_threshold, cmap=cmap
        )

        ax_im.spines[["top", "right"]].set_visible(False)

        if scalebar is not None:
            _plot_scalebar(ax_im, **scalebar)

        if histogram:
            vals, bins = np.histogram(
                integrity[signal > signal_threshold],
                bins=50,
                range=(0, 1),
                density=True,
            )
            colors = cmap(bins[1:-1])
            bars = ax_hist.barh(bins[1:-1], vals[1:], height=0.01)
            for i, bar in enumerate(bars):
                bar.set_color(colors[i])
            ax_hist.set(ylim=(0, 1), ylabel=VSI, xticks=[])
            ax_hist.yaxis.tick_right()
            ax_hist.yaxis.set_label_position("right")
            ax_hist.invert_xaxis()
            ax_hist.spines[["top", "bottom", "left"]].set_visible(False)

        else:
            fig.colorbar(img, label=VSI)

    return fig


def plot_region_of_interest(
    ovrlp: Ovrlp,
    x: float,
    y: float,
    *,
    window_size: int = 30,
    signal_threshold: float = _SIGNAL_THRESHOLD,
    scalebar: dict | None = SCALEBAR_PARAMS,
    figsize: tuple[float, float] = (12, 8),
    **kwargs,
) -> Figure:
    """
    Plot an overview of a zoomed-in region of interest.

    Parameters
    ----------
    ovrlp : Ovrlp
        Ovrlp object containing the fitted model
    x : float
        x coordinate of the region of interest
    y : float
        y coordinate of the region of interest
    window_size : int, optional
        Size of the window to display.
    signal_threshold : float, optional
        Threshold below which the signal is faded out in the VSI plot,
        to avoid displaying noisy areas with low predictive confidence.
    scalebar : dict[str, typing.Any] | None
        If `None` no scalebar will be plotted. Otherwise a dictionary with
        additional kwargs for ``matplotlib_scalebar.scalebar.ScaleBar``.
        By default :py:attr:`ovrlpy.SCALEBAR_PARAMS`
    figsize : tuple[float, float]
        Size of the figure in inches.
    kwargs
        Other keyword arguments are passed to :py:func:`matplotlib.pyplot.figure`

    Returns
    -------
    matplotlib.figure.Figure
    """
    integrity = ovrlp.integrity_map
    signal = ovrlp.signal_map
    embedding = ovrlp.pseudocells.obsm["2D_UMAP"]

    # first, create and color-embed the subsample of the region of interest
    roi_transcripts = ovrlp.subset_transcripts(x, y, window_size=window_size).sort("z")
    _, embedding_color = ovrlp.transform_transcripts(roi_transcripts)
    roi_transcripts = roi_transcripts.with_columns(RGB=embedding_color)

    roi = ((x - window_size, x + window_size), (y - window_size, y + window_size))

    fig = plt.figure(figsize=figsize, **kwargs)
    gs = fig.add_gridspec(2, 3)

    # integrity map
    ax_integrity = fig.add_subplot(gs[0, 2], label="signal_integrity", facecolor="k")

    img = _plot_signal_integrity(
        ax_integrity, integrity, signal, signal_threshold, cmap=BIH_CMAP
    )
    fig.colorbar(img, label=VSI)

    ax_integrity.set_title("ROI, signal integrity")
    ax_integrity.set_xlim(x - window_size, x + window_size)
    ax_integrity.set_ylim(y - window_size, y + window_size)

    # UMAP
    ax_umap = fig.add_subplot(gs[0, 0], label="umap")
    _plot_embeddings(ax_umap, embedding, ovrlp.pseudocells.obsm["RGB"])
    ax_umap.set_title("UMAP")

    # tissue map
    ax_tissue_whole: Axes = fig.add_subplot(gs[0, 1], label="celltype_map")
    plot_tissue(ovrlp, ax=ax_tissue_whole, s=1)

    roi_box = Rectangle(
        (x - window_size, y - window_size),
        2 * window_size,
        2 * window_size,
        fill=False,
        edgecolor="k",
        linewidth=2,
    )
    ax_tissue_whole.add_artist(roi_box)
    ax_tissue_whole.set_title("celltype map")

    # top view of ROI
    roi_scatter_kwargs = dict(marker=".", alpha=0.8, s=1.5e5 / window_size**2)

    ax_roi_top = fig.add_subplot(gs[1, 0], label="top_map")
    roi_top = roi_transcripts.filter(pl.col("z") > pl.col("z_center"))
    _plot_tissue_scatter(
        ax_roi_top,
        roi_top["x"],
        roi_top["y"],
        roi_top["RGB"].to_numpy(),
        title="ROI celltype map, top",
        **roi_scatter_kwargs,
    )
    _mark_roi_center(ax_roi_top, x, y, roi)

    # bottom view of ROI
    ax_roi_bottom = fig.add_subplot(gs[1, 1], label="bottom_map")
    roi_bottom = roi_transcripts.filter(pl.col("z") < pl.col("z_center"))[::-1]
    _plot_tissue_scatter(
        ax_roi_bottom,
        roi_bottom["x"],
        roi_bottom["y"],
        roi_bottom["RGB"].to_numpy(),
        title="ROI celltype map, bottom",
        **roi_scatter_kwargs,
    )
    _mark_roi_center(ax_roi_bottom, x, y, roi)

    if scalebar is not None:
        _plot_scalebar(ax_integrity, **scalebar)
        _plot_scalebar(ax_tissue_whole, **scalebar)
        _plot_scalebar(ax_roi_top, **scalebar)
        _plot_scalebar(ax_roi_bottom, **scalebar)

    # side view of ROI
    roi_side_scatter_kwargs = dict(s=10, alpha=0.5)

    sub_gs = gs[1, 2].subgridspec(2, 1)

    ax_side_x = fig.add_subplot(sub_gs[0, 0], label="x_cut")
    roi_side_x = roi_transcripts.filter(pl.col("y") < (y + 4), pl.col("y") > (y - 4))

    _plot_tissue_scatter(
        ax_side_x,
        roi_side_x["x"],
        roi_side_x["z"],
        roi_side_x["RGB"].to_numpy(),
        title="ROI, vertical, x-cut",
        **roi_side_scatter_kwargs,
    )

    ax_side_y = fig.add_subplot(sub_gs[1, 0], label="y_cut")
    roi_side_y = roi_transcripts.filter(pl.col("x") < (x + 4), pl.col("x") > (x - 4))

    _plot_tissue_scatter(
        ax_side_y,
        roi_side_y["y"],
        roi_side_y["z"],
        roi_side_y["RGB"].to_numpy(),
        title="ROI, vertical, y-cut",
        **roi_side_scatter_kwargs,
    )

    fig.tight_layout()

    return fig



import importlib.metadata
import os
import sys
from datetime import datetime

sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "ovrlpy"
copyright = f"""
{datetime.now():%Y}, Sebastian Tiesmeyer, Naveed Ishaque, Roland Eils,
Berlin Institute of Health @ Charité"""
author = "Sebastian Tiesmeyer"
version = importlib.metadata.version("ovrlpy")
release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration


extensions = [
    "sphinx_copybutton",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "autoapi.extension",
    "sphinx.ext.mathjax",
    "myst_nb",
]

nb_execution_mode = "off"


autodoc_typehints = "none"
autodoc_typehints_format = "short"

autoapi_dirs = ["../../ovrlpy"]
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "imported-members",
]
autoapi_own_page_level = "function"

python_use_unqualified_type_names = True  # still experimental

autosummary_generate = True
autosummary_imported_members = True

nitpicky = True
nitpick_ignore = [
    ("py:class", "numpy.typing.DTypeLike"),
    ("py:class", "polars.DataFrame"),
    ("py:class", "polars.DataType"),
    ("py:class", "umap.UMAP"),
    ("py:class", "optional"),
]

intersphinx_mapping = dict(
    anndata=("https://anndata.readthedocs.io/en/stable/", None),
    matplotlib=("https://matplotlib.org/stable/", None),
    numpy=("https://numpy.org/doc/stable/", None),
    pandas=("https://pandas.pydata.org/pandas-docs/stable/", None),
    polars=("https://docs.pola.rs/api/python/stable/", None),
    python=("https://docs.python.org/3", None),
    scipy=("https://docs.scipy.org/doc/scipy/", None),
    sklearn=("https://scikit-learn.org/stable/", None),
    umap=("https://umap-learn.readthedocs.io/page/", None),
)

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_logo = "../resources/ovrlpy-logo.png"
html_theme_options = {"logo_only": True}


def skip_attributes(app, what, name, obj, skip, options):
    if what == "attribute":
        skip = True
    return skip


def setup(sphinx):
    sphinx.connect("autoapi-skip-member", skip_attributes)



import warnings
from collections.abc import Callable, Sequence
from functools import reduce
from operator import add
from typing import Literal, TypeVar, get_args

import numpy as np
import polars as pl
from polars.datatypes import DataType, Float32, Int32
from scipy.sparse import coo_array


def _assign_xy(
    df: pl.DataFrame,
    xy_columns: Sequence[str] = ("x", "y"),
    gridsize: float = 1,
    shift: bool = True,
) -> pl.DataFrame:
    """
    Assigns an x,y pixel coordinate.

    Parameters
    ----------
    df : polars.DataFrame
        A dataframe of coordinates.
    xy_columns : list, optional
        The names of the columns containing the x,y-coordinates.
    gridsize : float, optional
        The size of the grid.
    shift : bool
        Whether to shift the x,y coordinates to 0 to minimize the analysis area.
        If False, the sample will keep its original shape.
    """
    lf = df.lazy()

    if shift:
        lf = lf.with_columns(pl.col(col) - pl.col(col).min() for col in xy_columns)
    elif df.select(pl.col(xy_columns).min() < 0).to_numpy().any():
        raise ValueError("shift=False is only possible if coordinates are > 0")

    lf = lf.with_columns(pl.col(xy_columns) / gridsize).with_columns(
        pl.col(xy_columns[0]).cast(Int32).shrink_dtype().alias("x_pixel"),
        pl.col(xy_columns[1]).cast(Int32).shrink_dtype().alias("y_pixel"),
    )

    return lf.collect(engine="streaming")  # reduce memory usage by using streaming


Shape = TypeVar("Shape", bound=tuple[int, ...])


def _message_passing(
    x: np.ndarray[Shape, np.dtype], /, n_iter: int
) -> np.ndarray[Shape, np.dtype]:
    with warnings.catch_warnings():
        # ignore 'mean of empty slice' warning if all values are nan
        warnings.simplefilter("ignore", category=RuntimeWarning)
        for _ in range(n_iter):
            x = reduce(
                add,
                (
                    np.nanmean([x, np.roll(x, shift, axis=axis)], axis=0)  # type: ignore
                    for axis in (0, 1)
                    for shift in (1, -1)
                ),
            )
            x /= 4
    return x


def _mean_elevation(
    df: pl.DataFrame, z_key: str, *, dtype: type[DataType]
) -> np.ndarray[tuple[int, int], np.dtype]:
    x = df["x_pixel"].to_numpy()
    y = df["y_pixel"].to_numpy()

    z = coo_array((df[z_key].cast(dtype).to_numpy(), (x, y))).toarray()
    n = coo_array((np.ones(len(df), dtype=np.uint16), (x, y))).toarray()

    with np.errstate(divide="ignore", invalid="ignore"):
        # ignore zero division warnings as we want to transform 0 to nan
        z /= n

    return z


def _assign_z_mean_message_passing(
    df: pl.DataFrame, n_iter: int, z_key: str = "z", *, dtype: type[DataType] = Float32
) -> pl.Series:
    elevation_map = _mean_elevation(df, z_key, dtype=dtype)
    elevation_map = _message_passing(elevation_map, n_iter)
    return pl.Series("z", elevation_map[df["x_pixel"], df["y_pixel"]])


def _transform(
    df: pl.DataFrame,
    f: Callable[[str], pl.Expr],
    z_key: str = "z",
    dtype: type[DataType] = Float32,
) -> pl.Series:
    # transforming over the on-the-fly created pixel-id is more efficient than
    # using the two pixel columns
    _pixel_id = pl.col("x_pixel") + pl.col("y_pixel") * (pl.col("x_pixel").max() + 1)
    z = df.lazy().select(f(z_key).over(_pixel_id).cast(dtype)).collect().to_series()
    return z


_METHODS = Literal["mean", "median", "message_passing"]


def process_coordinates(
    coordinates: pl.DataFrame,
    /,
    gridsize: float = 1,
    *,
    coordinate_keys: tuple[str, str, str] = ("x", "y", "z"),
    method: _METHODS = "message_passing",
    shift: bool = True,
    n_iter: int = 20,
    dtype: type[DataType] = Float32,
) -> pl.DataFrame:
    """
    Runs the pre-processing routine of the coordinate dataframe.

    x,y coordinates are rescaled by the gridsize and x,y pixels assigned to all
    molecules. A z-coordinate-center for each pixel is calculated.

    Parameters
    ----------
    coordinates : polars.DataFrame
        A dataframe of coordinates.
    gridsize : float, optional
        The size of the pixel grid.
    coordinate_keys : tuple[str, str, str], optional
        Name of the coordinate columns.
    method : bool, optional
        The measure to use to determine the z-dimension threshold for subslicing.
        One of, mean, median, and message_passing.
    shift : bool
        Whether to shift the x,y coordinates to 0 to minimize the analysis area.
        If False, the sample will keep its original shape.
    n_iter : int, optional
        Number of iterations. Only used if the method is message_passing.
    dtype : type[polars.DataType], optional
        Datatype of the z-coordinate center.

    Returns
    -------
    polars.DataFrame:
        A dataframe with added x_pixel, y_pixel, and z_center column.
    """
    if method not in get_args(_METHODS):
        raise ValueError(f"`method` must be one of {get_args(_METHODS)} not '{method}'")

    *xy, z = coordinate_keys

    coordinates = _assign_xy(coordinates, xy_columns=xy, gridsize=gridsize, shift=shift)

    if method == "message_passing":
        z_center = _assign_z_mean_message_passing(
            coordinates, n_iter=n_iter, z_key=z, dtype=dtype
        )
    else:
        if method == "mean":
            f = pl.mean
        elif method == "median":
            f = pl.median

        z_center = _transform(coordinates, f, z, dtype=dtype)

    return coordinates.with_columns(pl.Series("z_center", z_center))



import warnings
from collections.abc import Iterable
from queue import Empty, SimpleQueue
from typing import Any, Literal

import numpy as np
import polars as pl
from scipy.linalg import norm
from scipy.sparse import csr_array
from sklearn.decomposition import PCA
from sklearn.neighbors import radius_neighbors_graph
from umap import UMAP

from ._kde import find_local_maxima, kde_2d_discrete

UMAP_2D_PARAMS: dict[str, Any] = {"n_components": 2, "n_neighbors": 20, "min_dist": 0}
"""Default 2D-UMAP parameters"""

UMAP_RGB_PARAMS: dict[str, Any] = {"n_components": 3, "n_neighbors": 10, "min_dist": 0}
"""Default RGB-UMAP parameters"""


def _determine_localmax_and_sample(
    values: np.ndarray, min_distance: int = 3, min_value: float = 5
):
    """
    Returns a list of local maxima and their corresponding values.

    Parameters
    ----------
    values : np.ndarray
        A 2D array of values.
    min_distance : int, optional
        The minimum distance between local maxima.
    min_value : float, optional
        The minimum value to consider values as maxima.

    Returns
    -------
    rois_x
        x coordinates of local maxima.
    rois_y
        y coordinates of local maxima.
    values
        values at local maxima.
    """
    rois = find_local_maxima(values, min_distance, min_value)

    rois_x = rois[:, 0]
    rois_y = rois[:, 1]

    return rois_x, rois_y, values[rois_x, rois_y]


## These functions are going to be separated into a package of their own at some point:

# define a 45-degree 3D rotation matrix
_ROTATION_MATRIX = np.array(
    [
        [0.500, 0.500, -0.707],
        [-0.146, 0.854, 0.500],
        [0.854, -0.146, 0.500],
    ]
)


def _fill_color_axes(rgb, pca: PCA, *, fit: bool = False) -> np.ndarray:
    """rotate the transformed data 45° in all dimensions"""
    if fit:
        pca.fit(rgb)
    return np.dot(pca.transform(rgb), _ROTATION_MATRIX)


def _minmax_scaling(x: np.ndarray):
    """scale features (rows) to unit range"""
    x_min = x.min(axis=0)
    x_max = x.max(axis=0)
    return (x - x_min) / (x_max - x_min)


def _transform_embeddings(expression, pca: PCA, umap_2d: UMAP, umap_rgb: UMAP):
    """fit the expression data into the umap embeddings after PCA transformation"""
    factors = pca.transform(expression)

    embedding = umap_2d.transform(factors)
    embedding_color = umap_rgb.transform(factors / norm(factors, axis=1, keepdims=True))

    return embedding, embedding_color


def _gaussian_weighted_neighbors(
    coords, radius: float, bandwidth: float, *, n_workers: int | None = None
) -> csr_array:
    """gaussian-weighted radius-based neighbors graph"""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="X does not have valid feature names")
        neighbors = radius_neighbors_graph(
            coords, radius, mode="distance", include_self=True, n_jobs=n_workers
        )
    neighbors = csr_array(neighbors)
    neighbors.data = (1 / ((2 * np.pi) ** (3 / 2) * bandwidth**3)) * np.exp(
        -(neighbors.data**2) / (2 * bandwidth**2)
    )
    return neighbors


def _weighted_average_expression(
    genes: pl.Series,
    weights: np.ndarray | csr_array,
    gene_list: Iterable[str],
    normalize: bool = True,
) -> pl.DataFrame:
    """weighted average of the expression values of the neighbors"""

    assert weights.shape[0] == weights.shape[1] == len(genes)

    # drop name which makes having correct names after to_dummies easier
    genes = genes.rename("")
    # first cast to string for correct comparison
    genes = genes.cast(pl.String).cast(pl.Enum(gene_list), strict=False)
    # one-hot encoded genes
    genes_idx = genes.to_dummies(separator="").drop("null", strict=False)

    expression = pl.DataFrame(
        {gene: weights @ genes_idx[gene].to_numpy() for gene in genes_idx.columns}
    )

    if normalize:
        l2_norms = norm(expression.to_numpy(), axis=1)
        expression = expression.with_columns(pl.all().truediv(l2_norms))

    expression = expression.with_columns(
        pl.lit(0.0).alias(gene) for gene in gene_list if gene not in genes_idx.columns
    ).select(gene_list)

    return expression


def _gene_embedding(
    df: pl.DataFrame,
    mask: np.ndarray,
    factor: np.ndarray,
    xy: tuple[str, str] = ("x_pixel", "y_pixel"),
    **kwargs,
):
    """
    calculate top and bottom embedding

    Parameters
    ----------
    df : polars.DataFrame
        DataFrame of x, y, z, and z_center coordinates
    mask : numpy.ndarray
        binary mask for which pixels to calculate embedding
    factor : numpy.ndarray
        embedding weights
    """
    if len(df) < 2:
        return None, None

    x, y = xy

    # TODO: what happens if equal?
    top = df.select(xy).filter(df["z"] > df["z_center"])
    bottom = df.select(xy).filter(df["z"] < df["z_center"])

    if len(top) == 0:
        signal_top = None
    else:
        signal_top = kde_2d_discrete(
            top[x].to_numpy(), top[y].to_numpy(), size=mask.shape, **kwargs
        )[mask]
        signal_top = signal_top[:, None] * factor[None, :]

    if len(bottom) == 0:
        signal_bottom = None
    else:
        signal_bottom = kde_2d_discrete(
            bottom[x].to_numpy(), bottom[y].to_numpy(), size=mask.shape, **kwargs
        )[mask]
        signal_bottom = signal_bottom[:, None] * factor[None, :]

    return signal_top, signal_bottom


def _calculate_embedding(
    genes: SimpleQueue[tuple[int, pl.DataFrame]],
    mask: np.ndarray[tuple[int, int], np.dtype[np.bool_]],
    components: np.ndarray[tuple[int, int], np.dtype[np.floating]],
    **kwargs,
) -> tuple[np.ndarray | Literal[0], np.ndarray | Literal[0]]:
    embedding_top: Literal[0] | np.ndarray = 0
    embedding_bottom: Literal[0] | np.ndarray = 0

    while True:
        try:
            i, gene = genes.get(block=False)
        except Empty:
            break

        top, bottom = _gene_embedding(gene, mask, components[:, i], **kwargs)
        if top is not None:
            embedding_top += top
        if bottom is not None:
            embedding_bottom += bottom

    return embedding_top, embedding_bottom


def _cosine_similarity(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    norm_ = norm(x, axis=1) * norm(y, axis=1)
    norm_[norm_ == 0] = np.inf
    return np.sum(x * y, axis=1) / norm_
